"""Structures for auto-completion and data storage."""
from dataclasses import dataclass

from olli.config import TokenConfig


@dataclass
class TokenMatch:
    """
    A dataclass representing a searched token.

    The token is an instance of TokenConfig which contains the information on which token
    was found. The services dictionary contains a map of service names to the number of times
    this specific token appeared in their logs.
    """

    token: TokenConfig
    services: dict[str, int]



"""This module contains the logic for sending webhooks to Discord."""
from datetime import datetime

import httpx
from loguru import logger

from olli.config import CONFIG
from olli.structures import TokenMatch


def send_olli_error(error: str) -> None:
    """Send an error embed containing the passed error message to Discord."""
    logger.info("Sending error payload to Discord")
    httpx.post(CONFIG.discord.webhook_url, json={
        "embeds": [{
            "title": "Olli Error",
            "color": 0xff5f5f,
            "description": f"Olli encountered an error: {error}",
            "timestamp": datetime.utcnow().isoformat()
        }]
    })


def send_token_matches(matches: list[TokenMatch]) -> None:
    """Send the found token matches to Discord.

    The method checks that each token has at least 1 match and then merges
    together up to 10 embeds into one webhook payload.
    """
    embeds = []

    for match in matches:
        if sum(match.services.values()) == 0:
            continue

        embed = {
            "title": f"Logs - {match.token.token}",
            "description": f"`{sum(match.services.values())}` events matching",
            "color": int(match.token.color[1:], 16),
            "author": {
                "name": "Olli"
            },
            "footer": {
                "text": f"Last {CONFIG.olli.interval_minutes} minutes"
            },
            "timestamp": datetime.utcnow().isoformat(),
            "fields": []
        }

        for service, count in match.services.items():
            embed["fields"].append({
                "name": service,
                "value": f"`{count:,}` logs",
                "inline": True
            })

        embeds.append(embed)

    if len(embeds) > 0:
        logger.info("Sending alerts payload to Discord")
        resp = httpx.post(CONFIG.discord.webhook_url, json={
            "embeds": embeds
        })

        resp.raise_for_status()
    else:
        logger.info("No alerts to send to Discord")



"""Configuration loading and validation."""
import os
from pathlib import Path
from typing import Optional

import toml
from dotenv import load_dotenv
from loguru import logger
from pydantic import BaseModel, validator

load_dotenv()


# All the locations where we might find a config file
CONFIG_PATHS = [
    "./olli.toml",
    "/config/olli.toml",
    "/olli/config.toml",
    "/etc/olli/config.toml"
]


class TokenConfig(BaseModel):
    """Class representing a token config entry."""

    token: str
    color: str = "#7289DA"


class LokiConfig(BaseModel):
    """Loki specific configuration."""

    api_url: str
    jobs: list[str]
    max_logs: Optional[int] = 5_000


class DiscordConfig(BaseModel):
    """Configuration for Discord alerting."""

    webhook_url: Optional[str]

    @validator("webhook_url", always=True)
    def env_provided_webhook(cls, value: Optional[str]) -> str:
        """
        If no webhook is specified in the config, try fetch from environment.

        If not found in the environment either then raise a validation error.
        """
        if value:
            return value

        if webhook := os.environ.get("WEBHOOK_URL"):
            return webhook

        raise ValueError(
            "Must specify webhook_url under [discord] or WEBHOOK_URL env var"
        )


class ServiceConfig(BaseModel):
    """Configuration of the Olli status."""

    interval_minutes: int
    tokens: list[TokenConfig]

    @validator("interval_minutes")
    def must_be_above_zero(cls, value: int) -> int:
        """Validate that the interval minutes is 1 or greater."""
        if value < 1:
            raise ValueError("Interval must be above zero minutes.")

        return value


class OlliConfig(BaseModel):
    """Class representing root Olli config."""

    loki: LokiConfig
    olli: ServiceConfig
    discord: DiscordConfig = DiscordConfig()


def get_config() -> OlliConfig:
    """Open the config file, parse the TOML and convert to Pydantic objects."""
    logger.info("Searching for config file")

    path = None

    for file_path in CONFIG_PATHS:
        if (config_file := Path(file_path)).exists():
            path = config_file
            logger.info(f"Found config at {file_path}")
            break

    if not path:
        logger.critical("Could not find a config file. Please refer to the documentation.")
        raise SystemExit(1)

    with open(config_file) as conf_file:
        return OlliConfig(**toml.load(conf_file))


CONFIG = get_config()



"""This module contains the logic for grepping logs and generating Discord alerts."""
import httpx
from loguru import logger

from olli import webhook
from olli.api import LokiHTTPClient
from olli.config import CONFIG, TokenConfig
from olli.structures import TokenMatch

api_client = LokiHTTPClient()


def get_match(token: TokenConfig) -> TokenMatch:
    """Search the configured service logs for a given token."""
    try:
        logger.debug(f"Searching for token {token.token}")
        svc_logs = api_client.get_token_logs(token.token)
    except httpx.ConnectError:
        logger.error("Could not connect to Loki")
        return webhook.send_olli_error("Loki refused to connect.")
    except httpx.HTTPStatusError as e:
        logger.error(f"Loki returned error status code: {e.response.status_code}")
        return webhook.send_olli_error(
            f"Loki returned a bad status code: `{e.response.status_code}`"
        )

    if svc_logs["status"] != "success":
        logger.error(f"Received an error response from Loki for token {token.token}")
        return webhook.send_olli_error(
            "Loki returned an error, check your service names are correct."
        )

    match = TokenMatch(token, {})

    for service in svc_logs["data"]["result"]:
        app = service["stream"].get("app")
        svc_name = app if app else service["stream"].get("job")

        count = len(service["values"])

        match.services[svc_name] = count

    logger.debug(f"Finished searching for {token.token}")

    return match


def send_alerts(matches: list[TokenMatch]) -> None:
    """Take the found matches and relay them to Discord."""
    # TODO: batch into 10 to avoid embed limits
    logger.info("Sending webhook payload to Discord")
    webhook.send_token_matches(matches)


def run() -> None:
    """Entrypoint for Olli's search process."""
    logger.info("Running Olli search")
    matches = []

    for token in CONFIG.olli.tokens:
        matches.append(get_match(token))

    send_alerts(matches)
    logger.info("Olli search complete.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['olli']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.17.1,<0.18.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.8.1,<2.0.0',
 'python-dotenv>=0.16.0,<0.17.0',
 'schedule>=1.0.0,<2.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['olli = olli.__main__:start']}

setup_kwargs = {
    'name': 'olli',
    'version': '0.0.1',
    'description': 'Olli searches your Loki logs and relays matching terms to Discord.',
    'long_description': '# Olli\n\nOlli searches your Loki logs and relays matching terms to Discord. Please read the [documentation](https://python-discord.github.io/olli/) to get started.\n\n- [Documentation](https://python-discord.github.io/olli/)\n- [Docker Image](https://ghcr.io/python-discord/olli)\n\n## Contributing\n\nPlease open an issue before contributing features to discuss implementation.\n\n## License\n\nOlli is licensed under MIT.\n',
    'author': 'Joe Banks',
    'author_email': 'joseph@josephbanks.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://python-discord.github.io/olli/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
